# Copyright (c) 2018 The Regents of the University of Michigan
# All rights reserved.
# This software is licensed under the BSD 3-Clause License.
import os
import errno
import json
import argparse
import logging
from contextlib import contextmanager

from signac.common import six


def _mkdir_p(path):
    """"Create a directory at 'path', ignore if the directory already exists.

    Needed, because the 'ignore_exists' is not available in Python 2.7.
    """
    try:
        os.makedirs(path)
    except OSError as error:
        if not (error.errno == errno.EEXIST and os.path.isdir(path)):
            raise


def _positive_int(value):
    """Expect a command line argument to be a positive integer.

    Designed to be used in conjunction with an argparse.ArgumentParser.

    :param value:
        This function will raise an argparse.ArgumentTypeError if value
        is not a positive integer.
    :raises:
        :class:`argparse.ArgumentTypeError`
    """
    try:
        ivalue = int(value)
        if ivalue <= 0:
            raise argparse.ArgumentTypeError("Value must be positive.")
    except (TypeError, ValueError):
        raise argparse.ArgumentTypeError(
            "{} must be a positive integer.".format(value))
    return ivalue


def draw_progressbar(value, total, width=40):
    """Visualize progess with a progress bar.

    :param value:
        The current progress as a fraction of total.
    :type value:
        int
    :param total:
        The maximum value that 'value' may obtain.
    :type total:
        int
    :param width:
        The character width of the drawn progress bar.
    :type width:
        int
    """
    "Helper function for the visualization of progress."
    assert value >= 0 and total > 0
    n = int(value / total * width)
    return '|' + ''.join(['#'] * n) + ''.join(['-'] * (width - n)) + '|'


def write_human_readable_statepoint(script, job):
    """Human-readable representation of a signac state point."""
    script.write('# Statepoint:\n#\n')
    sp_dump = json.dumps(job.statepoint(), indent=2).replace(
        '{', '{{').replace('}', '}}')
    for line in sp_dump.splitlines():
        script.write('# ' + line + '\n')


@contextmanager
def redirect_log(job, filename='run.log', formatter=None, logger=None):
    """Redirect all messages logged via the logging interface to the given file.

    :param job:
        An instance of a signac job.
    :type job:
        :class:`signac.Project.Job`
    :formatter:
        The logging formatter to use, uses a default formatter if this argument
        is not provided.
    :type formatter:
        :class:`logging.Formatter`
    :param logger:
        The instance of logger to which the new file log handler is added. Defaults
        to the default logger returned by `logging.getLogger()` if this argument is
        not provided.
    type logger:
        :class:`logging.Logger`
    """
    if formatter is None:
        formatter = logging.Formatter(
            '%(asctime)s %(name)-12s %(levelname)-8s %(message)s')
    if logger is None:
        logger = logging.getLogger()

    filehandler = logging.FileHandler(filename=job.fn('run.log'))
    filehandler.setFormatter(formatter)
    logger.addHandler(filehandler)
    try:
        yield
    finally:
        logger.removeHandler(filehandler)


@contextmanager
def add_path_to_environment_pythonpath(path):
    "Temporarily insert the current working directory into the environment PYTHONPATH variable."
    path = os.path.realpath(path)
    pythonpath = os.environ.get('PYTHONPATH')
    if pythonpath:
        for path_ in pythonpath:
            if os.path.isabs(path_) and os.path.realpath(path_) == path:
                yield   # Path is already in PYTHONPATH, nothing to do here.
                return
        try:
            # Append the current working directory to the PYTHONPATH.
            tmp_path = [path] + pythonpath.split(':')
            os.environ['PYTHONPATH'] = ':'.join(tmp_path)
            yield
        finally:
            os.environ['PYTHONPATH'] = pythonpath
            pass
    else:
        try:
            # The PYTHONPATH was previously not set, set to current working directory.
            os.environ['PYTHONPATH'] = path
            yield
        finally:
            del os.environ['PYTHONPATH']


@contextmanager
def add_cwd_to_environment_pythonpath():
    with add_path_to_environment_pythonpath(os.getcwd()):
        yield


@contextmanager
def switch_to_directory(root=None):
    "Temporarily switch into the given root directory (if not None)."
    if root is None:
        yield
    else:
        cwd = os.getcwd()
        try:
            os.chdir(root)
            yield
        finally:
            os.chdir(cwd)


def _is_identifier(name):
    """Check if 'name' is a valid Python identifier.

    Source: https://stackoverflow.com/a/2545164
    """
    if name:
        if six.PY2:
            import re
            import keyword
            return re.match(r'^[a-z_][a-z0-9_]*$', name, re.I) and not keyword.iskeyword(name)
        else:
            return name.isidentifier()
    else:
        return False    # empty string or None



# Copyright (c) 2018 The Regents of the University of Michigan
# All rights reserved.
# This software is licensed under the BSD 3-Clause License.
import os
import subprocess
import tempfile
import json

from .base import Scheduler
from .base import ClusterJob, JobStatus


class SimpleScheduler(Scheduler):

    @classmethod
    def is_present(cls):
        return bool(os.environ.get('SIMPLE_SCHEDULER'))

    def __init__(self):
        self.cmd = os.environ['SIMPLE_SCHEDULER'].split()

    def jobs(self):
        cmd = self.cmd + ['status', '--json']
        status = json.loads(subprocess.check_output(cmd).decode('utf-8'))
        for _id, doc in status.items():
            yield ClusterJob(doc['job_name'], JobStatus(doc['status']))

    def submit(self, script, pretend=False, **kwargs):
        cmd = self.cmd + ['submit']
        if pretend:
            print("# Submit command: {}".format(' '.join(cmd)))
            print(script)
            print()
        else:
            with tempfile.NamedTemporaryFile() as tmp_submit_script:
                tmp_submit_script.write(str(script).encode('utf-8'))
                tmp_submit_script.flush()
                subprocess.check_output(cmd + [tmp_submit_script.name])
                return True



#!C:\bld\signac-flow_1538815040027\_h_env\python.exe
# EASY-INSTALL-ENTRY-SCRIPT: 'signac-flow==0.6.3','console_scripts','flow'
__requires__ = 'signac-flow==0.6.3'
import re
import sys
from pkg_resources import load_entry_point

if __name__ == '__main__':
    sys.argv[0] = re.sub(r'(-script\.pyw?|\.exe)?$', '', sys.argv[0])
    sys.exit(
        load_entry_point('signac-flow==0.6.3', 'console_scripts', 'flow')()
    )



# Copyright (c) 2018 The Regents of the University of Michigan
# All rights reserved.
# This software is licensed under the BSD 3-Clause License.
"""Contains logic for working with flow related information
stored in the signac config"""

from signac.common import config

from ..errors import ConfigKeyError


class _GetConfigValueNoneType(object):
    pass


_GET_CONFIG_VALUE_NONE = _GetConfigValueNoneType()


MISSING_ENV_CONF_KEY_MSG = """Your environment is missing the following configuration key: '{key}'
Please provide the missing information, for example by adding it to your global configuration:

signac config --global set {key} <VALUE>
"""


def require_config_value(key, ns=None, default=_GET_CONFIG_VALUE_NONE):
    """Request a value from the user's configuration, fail if not available.

    This method should be used whenever values need to be provided
    that are specific to a users's environment. A good example are
    account names.

    When a key is not configured and no default value is provided,
    a :py:class:`~.errors.SubmitError` will be raised and the user will
    be prompted to add the missing key to their configuration.

    Please note, that the key will be automatically expanded to
    be specific to this environment definition. For example, a
    key should be 'account', not 'MyEnvironment.account`.

    :param key: The environment specific configuration key.
    :type key: str
    :param ns: The namespace in which to look for the key
    :type key: str
    :param default: A default value in case the key cannot be found
        within the user's configuration.
    :return: The value or default value.
    :raises ConfigKeyError: If the key is not in the user's configuration
        and no default value is provided.
    """
    try:
        if ns is None:
            return config.load_config()['flow'][key]
        else:
            return config.load_config()['flow'][ns][key]
    except KeyError:
        if default is _GET_CONFIG_VALUE_NONE:
            k = str(key) if ns is None else '{}.{}'.format(ns, key)
            print(MISSING_ENV_CONF_KEY_MSG.format(key='flow.' + k))
            raise ConfigKeyError("Missing environment configuration key: '{}'".format(k))
        else:
            return default


def get_config_value(key, ns=None, default=None):
    """Request a value from the user's configuration.

    This method should be used whenever values need to be provided
    that are specific to a users's environment. A good example are
    account names.

    When a key is not configured and no default value is provided,
    a :py:class:`~.errors.SubmitError` will be raised and the user will
    be prompted to add the missing key to their configuration.

    Please note, that the key will be automatically expanded to
    be specific to this environment definition. For example, a
    key should be 'account', not 'MyEnvironment.account`.

    :param key: The environment specific configuration key.
    :type key: str
    :param ns: The namespace in which to look for the key
    :type key: str
    :param default: A default value in case the key cannot be found
        within the user's configuration.
    :return: The value if found, None if not found.
    """
    try:
        if ns is None:
            return config.load_config()['flow'][key]
        else:
            return config.load_config()['flow'][ns][key]
    except KeyError:
        return default



# Copyright (c) 2018 The Regents of the University of Michigan
# All rights reserved.
# This software is licensed under the BSD 3-Clause License.
"""Implementation of the scheduling system for TORQUE schedulers.

This module implements the Scheduler class, and the JobCluster class
for TORQUE schedulers.
"""
from __future__ import print_function
import io
import getpass
import subprocess
import tempfile
import logging
import xml.etree.ElementTree as ET

from .base import Scheduler
from .base import ClusterJob, JobStatus


logger = logging.getLogger(__name__)


def _fetch(user=None):
    "Fetch the cluster job status information from the TORQUE scheduler."
    if user is None:
        user = getpass.getuser()
    cmd = "qstat -fx -u {user}".format(user=user)
    try:
        result = io.BytesIO(subprocess.check_output(cmd.split()))
    except FileNotFoundError:
        raise RuntimeError("Torque not available.")
    tree = ET.parse(source=result)
    return tree.getroot()


class TorqueJob(ClusterJob):
    "Implementation of the abstract ClusterJob class for TORQUE schedulers."

    def __init__(self, node):
        self.node = node

    def _id(self):
        return self.node.find('Job_Id').text

    def __str__(self):
        return str(self._id())

    def name(self):
        return self.node.find('Job_Name').text

    def status(self):
        job_state = self.node.find('job_state').text
        if job_state == 'R':
            return JobStatus.active
        if job_state == 'Q':
            return JobStatus.queued
        if job_state == 'C':
            return JobStatus.inactive
        if job_state == 'H':
            return JobStatus.held
        return JobStatus.registered


class TorqueScheduler(Scheduler):
    """Implementation of the abstract Scheduler class for TORQUE schedulers.

    This class allows us to submit cluster jobs to a TORQUE scheduler and query
    their current status.

    :param user:
        Limit the status information to cluster jobs submitted by user.
    :type user:
        str
    """
    # The standard command used to submit jobs to the TORQUE scheduler.
    submit_cmd = ['qsub']

    def __init__(self, user=None, **kwargs):
        super(TorqueScheduler, self).__init__(**kwargs)
        self.user = user

    def jobs(self):
        "Yield cluster jobs by querying the scheduler."
        self._prevent_dos()
        nodes = _fetch(user=self.user)
        for node in nodes.findall('Job'):
            yield TorqueJob(node)

    def submit(self, script, after=None, pretend=False, hold=False, flags=None, *args, **kwargs):
        """Submit a job script for execution to the scheduler.

        :param script:
            The job script submitted for execution.
        :type script:
            str
        :param after:
            Execute the submitted script after a job with this id has completed.
        :type after:
            str
        :param pretend:
            If True, do not actually submit the script, but only simulate the submission.
            Can be used to test whether the submission would be successful.
            Please note: A successful "pretend" submission is not guaranteed to succeeed.
        :type pretend:
            bool
        :param flags:
            Additional arguments to pass through to the scheduler submission command.
        :type flags:
            list
        :returns:
            The cluster job id if the script was successfully submitted, otherwise None.
        """
        if flags is None:
            flags = []
        elif isinstance(flags, str):
            flags = flags.split()

        submit_cmd = self.submit_cmd + flags

        if after is not None:
            submit_cmd.extend(
                ['-W', 'depend="afterok:{}"'.format(after.split('.')[0])])

        if hold:
            submit_cmd += ['-h']

        if pretend:
            print("# Submit command: {}".format(' '.join(submit_cmd)))
            print(script)
            print()
        else:
            with tempfile.NamedTemporaryFile() as tmp_submit_script:
                tmp_submit_script.write(str(script).encode('utf-8'))
                tmp_submit_script.flush()
                output = subprocess.check_output(
                    submit_cmd + [tmp_submit_script.name])
            jobsid = output.decode('utf-8').strip()
            return jobsid

    @classmethod
    def is_present(cls):
        "Return True if it appears that a TORQUE scheduler is available within the environment."
        try:
            subprocess.check_output(['qsub', '--version'], stderr=subprocess.STDOUT)
        except (IOError, OSError):
            return False
        else:
            return True



# Copyright (c) 2017 The Regents of the University of Michigan
# All rights reserved.
# This software is licensed under the BSD 3-Clause License.
"""Environments for incite supercomputers."""
from ..environment import DefaultTorqueEnvironment

import math


class TitanEnvironment(DefaultTorqueEnvironment):
    """Environment profile for the titan super computer.

    https://www.olcf.ornl.gov/titan/
    """
    hostname_pattern = 'titan'
    template = 'titan.sh'
    cores_per_node = 16

    @classmethod
    def mpi_cmd(cls, cmd, np):
        return "aprun -n {np} -N 1 -b {cmd}".format(cmd=cmd, np=np)

    @classmethod
    def gen_tasks(cls, js, np_total):
        """Helper function to generate the number of tasks (for overriding)"""
        js.writeline('#PBS -l nodes={}'.format(math.ceil(np_total/cls.cores_per_node)))
        return js

    @classmethod
    def script(cls, _id, **kwargs):
        js = super(TitanEnvironment, cls).script(_id=_id, **kwargs)
        js.writeline('#PBS -A {}'.format(cls.get_config_value('account')))
        return js


class EosEnvironment(DefaultTorqueEnvironment):
    """Environment profile for the eos super computer.

    https://www.olcf.ornl.gov/computing-resources/eos/
    """
    hostname_pattern = 'eos'
    template = 'eos.sh'
    cores_per_node = 32

    @classmethod
    def mpi_cmd(cls, cmd, np):
        return "aprun -n {np} -b {cmd}".format(cmd=cmd, np=np)

    @classmethod
    def gen_tasks(cls, js, np_total):
        """Helper function to generate the number of tasks (for overriding)"""
        js.writeline('#PBS -l nodes={}'.format(math.ceil(np_total/cls.cores_per_node)))
        return js

    @classmethod
    def script(cls, _id, **kwargs):
        js = super(EosEnvironment, cls).script(_id=_id, **kwargs)
        js.writeline('#PBS -A {}'.format(cls.get_config_value('account')))
        return js


__all__ = ['TitanEnvironment', 'EosEnvironment']



# Copyright (c) 2018 The Regents of the University of Michigan
# All rights reserved.
# This software is licensed under the BSD 3-Clause License.
"""Environments for the University of Michigan HPC environment."""
from ..environment import DefaultTorqueEnvironment

import math


class FluxEnvironment(DefaultTorqueEnvironment):
    """Environment profile for the flux supercomputing environment.

    http://arc-ts.umich.edu/systems-and-services/flux/
    """
    hostname_pattern = '(nyx|flux).*.umich.edu'
    template = 'umich.sh'
    cores_per_node = 1

    @classmethod
    def mpi_cmd(cls, cmd, np):
        return "mpirun -np {np} {cmd}".format(cmd=cmd, np=np)

    @classmethod
    def gen_tasks(cls, js, np_total, mode):
        """Helper function to generate the number of tasks (for overriding)"""
        if mode == 'cpu':
            js = js.writeline('#PBS -l nodes={}'.format(
                math.ceil(np_total/cls.cores_per_node)))
        elif mode == 'gpu':
            js.writeline('#PBS -l nodes={np}:gpus=1'.format(
                np=math.ceil(np_total/cls.cores_per_node)))
        return js

    @classmethod
    def script(cls, _id, np_total, mode, memory, **kwargs):
        js = super(FluxEnvironment, cls).script(_id=_id, **kwargs)
        js.writeline('#PBS -A {}'.format(cls.get_config_value('account')))
        js.writeline('#PBS -l qos={}'.format(cls.get_config_value('qos', 'flux')))
        js.writeline('#PBS -l pmem={}'.format(memory))
        if mode == 'cpu':
            js.writeline('#PBS -q {}'.format(cls.get_config_value('cpu_queue', 'flux')))
            js = cls.gen_tasks(js, np_total, mode)
        elif mode == 'gpu':
            q = cls.get_config_value('gpu_queue', cls.get_config_value('cpu_queue', 'flux') + 'g')
            js.writeline('#PBS -q {}'.format(q))
            js = cls.gen_tasks(js, np_total, mode)
        else:
            raise ValueError("Unknown mode '{}'.".format(mode))
        return js

    @classmethod
    def add_args(cls, parser):
        super(FluxEnvironment, cls).add_args(parser)
        parser.add_argument(
            '--mode',
            choices=('cpu', 'gpu'),
            default='cpu',
            help="Specify whether to submit to the CPU or the GPU queue. "
                 "(default=cpu)")
        parser.add_argument(
            '--memory',
            default='4g',
            help="Specify how much memory to reserve per node. (default=4g)")


__all__ = ['FluxEnvironment']



# Copyright (c) 2018 The Regents of the University of Michigan
# All rights reserved.
# This software is licensed under the BSD 3-Clause License.
"""Definition of base classes for the scheduling system."""
import enum
import time
import warnings


class JobStatus(enum.IntEnum):
    """Classifies the job's execution status."""
    unknown = 1
    registered = 2
    inactive = 3
    submitted = 4
    held = 5
    queued = 6
    active = 7
    error = 8
    # All user statuses are >= 128.
    user = 128


class ClusterJob(object):
    """This class represents a cluster job."""

    def __init__(self, jobid, status=None):
        self._job_id = jobid
        self._status = status

    def _id(self):
        return self._job_id

    def __str__(self):
        return str(self._id())

    def name(self):
        return self._id()

    def status(self):
        return self._status


class Scheduler(object):
    """Abstract base class for schedulers."""

    def __init__(self, header=None, cores_per_node=None):
        if header is not None:
            warnings.warn(
                "Scheduler: Ignoring deprecated 'header' argument.", DeprecationWarning)
        if cores_per_node is not None:
            warnings.warn(
                "Scheduler: Ignoring deprecated 'cores_per_node' argument.", DeprecationWarning)

    # The UNIX time stamp of the last scheduler query.
    _last_query = None

    # The amount of time in seconds a user needs to wait, before we
    # assume that repeated scheduler queries might risk a denial-of-service attack.
    _dos_timeout = 10

    @classmethod
    def _prevent_dos(cls):
        """This method should be called before querying the scheduler.

        If this method will raise an exception if it is called more than
        once within a time window defined by the '_dos_timeout' class.
        This is to prevent an (accidental) denial-of-service attack on
        the scheduling system.
        """
        if cls._last_query is not None:
            if time.time() - cls._last_query < cls._dos_timeout:
                raise RuntimeError(
                    "Too many scheduler requests within a short time!")
        cls._last_query = time.time()

    def jobs(self):
        """Yield all cluster jobs.

        :yields:
            :class:`.ClusterJob`
        """
        raise NotImplementedError()



"""
GUI progressbar decorator for iterators.
Includes a default (x)range iterator printing to stderr.

Usage:
  >>> from tqdm_gui import tgrange[, tqdm_gui]
  >>> for i in tgrange(10): #same as: for i in tqdm_gui(xrange(10))
  ...     ...
"""
# future division is important to divide integers and get as
# a result precise floating numbers (instead of truncated int)
from __future__ import division, absolute_import
# import compatibility functions and utilities
import sys
from time import time
from ._utils import _range
# to inherit from the tqdm class
from ._tqdm import tqdm


__author__ = {"github.com/": ["casperdcl", "lrq3000"]}
__all__ = ['tqdm_gui', 'tgrange']


class tqdm_gui(tqdm):  # pragma: no cover
    """
    Experimental GUI version of tqdm!
    """

    @classmethod
    def write(cls, s, file=sys.stdout, end="\n"):
        """
        Print a message via tqdm_gui (just an alias for print)
        """
        # TODO: print text on GUI?
        file.write(s)
        file.write(end)

    def __init__(self, *args, **kwargs):
        import matplotlib as mpl
        import matplotlib.pyplot as plt
        from collections import deque
        kwargs['gui'] = True

        super(tqdm_gui, self).__init__(*args, **kwargs)

        # Initialize the GUI display
        if self.disable or not kwargs['gui']:
            return

        self.fp.write('Warning: GUI is experimental/alpha\n')
        self.mpl = mpl
        self.plt = plt
        self.sp = None

        # Remember if external environment uses toolbars
        self.toolbar = self.mpl.rcParams['toolbar']
        self.mpl.rcParams['toolbar'] = 'None'

        self.mininterval = max(self.mininterval, 0.5)
        self.fig, ax = plt.subplots(figsize=(9, 2.2))
        # self.fig.subplots_adjust(bottom=0.2)
        if self.total:
            self.xdata = []
            self.ydata = []
            self.zdata = []
        else:
            self.xdata = deque([])
            self.ydata = deque([])
            self.zdata = deque([])
        self.line1, = ax.plot(self.xdata, self.ydata, color='b')
        self.line2, = ax.plot(self.xdata, self.zdata, color='k')
        ax.set_ylim(0, 0.001)
        if self.total:
            ax.set_xlim(0, 100)
            ax.set_xlabel('percent')
            self.fig.legend((self.line1, self.line2), ('cur', 'est'),
                            loc='center right')
            # progressbar
            self.hspan = plt.axhspan(0, 0.001,
                                     xmin=0, xmax=0, color='g')
        else:
            # ax.set_xlim(-60, 0)
            ax.set_xlim(0, 60)
            ax.invert_xaxis()
            ax.set_xlabel('seconds')
            ax.legend(('cur', 'est'), loc='lower left')
        ax.grid()
        # ax.set_xlabel('seconds')
        ax.set_ylabel((self.unit if self.unit else 'it') + '/s')
        if self.unit_scale:
            plt.ticklabel_format(style='sci', axis='y',
                                 scilimits=(0, 0))
            ax.yaxis.get_offset_text().set_x(-0.15)

        # Remember if external environment is interactive
        self.wasion = plt.isinteractive()
        plt.ion()
        self.ax = ax

    def __iter__(self):
        # TODO: somehow allow the following:
        # if not self.gui:
        #   return super(tqdm_gui, self).__iter__()
        iterable = self.iterable
        if self.disable:
            for obj in iterable:
                yield obj
            return

        # ncols = self.ncols
        mininterval = self.mininterval
        maxinterval = self.maxinterval
        miniters = self.miniters
        dynamic_miniters = self.dynamic_miniters
        unit = self.unit
        unit_scale = self.unit_scale
        ascii = self.ascii
        start_t = self.start_t
        last_print_t = self.last_print_t
        last_print_n = self.last_print_n
        n = self.n
        # dynamic_ncols = self.dynamic_ncols
        smoothing = self.smoothing
        avg_time = self.avg_time
        bar_format = self.bar_format

        plt = self.plt
        ax = self.ax
        xdata = self.xdata
        ydata = self.ydata
        zdata = self.zdata
        line1 = self.line1
        line2 = self.line2

        for obj in iterable:
            yield obj
            # Update and print the progressbar.
            # Note: does not call self.update(1) for speed optimisation.
            n += 1
            delta_it = n - last_print_n
            # check the counter first (avoid calls to time())
            if delta_it >= miniters:
                cur_t = time()
                delta_t = cur_t - last_print_t
                if delta_t >= mininterval:
                    elapsed = cur_t - start_t
                    # EMA (not just overall average)
                    if smoothing and delta_t:
                        avg_time = delta_t / delta_it \
                            if avg_time is None \
                            else smoothing * delta_t / delta_it + \
                            (1 - smoothing) * avg_time

                    # Inline due to multiple calls
                    total = self.total
                    # instantaneous rate
                    y = delta_it / delta_t
                    # overall rate
                    z = n / elapsed
                    # update line data
                    xdata.append(n * 100.0 / total if total else cur_t)
                    ydata.append(y)
                    zdata.append(z)

                    # Discard old values
                    # xmin, xmax = ax.get_xlim()
                    # if (not total) and elapsed > xmin * 1.1:
                    if (not total) and elapsed > 66:
                        xdata.popleft()
                        ydata.popleft()
                        zdata.popleft()

                    ymin, ymax = ax.get_ylim()
                    if y > ymax or z > ymax:
                        ymax = 1.1 * y
                        ax.set_ylim(ymin, ymax)
                        ax.figure.canvas.draw()

                    if total:
                        line1.set_data(xdata, ydata)
                        line2.set_data(xdata, zdata)
                        try:
                            poly_lims = self.hspan.get_xy()
                        except AttributeError:
                            self.hspan = plt.axhspan(0, 0.001, xmin=0,
                                                     xmax=0, color='g')
                            poly_lims = self.hspan.get_xy()
                        poly_lims[0, 1] = ymin
                        poly_lims[1, 1] = ymax
                        poly_lims[2] = [n / total, ymax]
                        poly_lims[3] = [poly_lims[2, 0], ymin]
                        if len(poly_lims) > 4:
                            poly_lims[4, 1] = ymin
                        self.hspan.set_xy(poly_lims)
                    else:
                        t_ago = [cur_t - i for i in xdata]
                        line1.set_data(t_ago, ydata)
                        line2.set_data(t_ago, zdata)

                    ax.set_title(self.format_meter(
                        n, total, elapsed, 0,
                        self.desc, ascii, unit, unit_scale,
                        1 / avg_time if avg_time else None, bar_format),
                        fontname="DejaVu Sans Mono", fontsize=11)
                    plt.pause(1e-9)

                    # If no `miniters` was specified, adjust automatically
                    # to the maximum iteration rate seen so far.
                    if dynamic_miniters:
                        if maxinterval and delta_t > maxinterval:
                            # Set miniters to correspond to maxinterval
                            miniters = delta_it * maxinterval / delta_t
                        elif mininterval and delta_t:
                            # EMA-weight miniters to converge
                            # towards the timeframe of mininterval
                            miniters = smoothing * delta_it * mininterval \
                                / delta_t + (1 - smoothing) * miniters
                        else:
                            miniters = smoothing * delta_it + \
                                (1 - smoothing) * miniters

                    # Store old values for next call
                    last_print_n = n
                    last_print_t = cur_t

        # Closing the progress bar.
        # Update some internal variables for close().
        self.last_print_n = last_print_n
        self.n = n
        self.close()

    def update(self, n=1):
        # if not self.gui:
        #   return super(tqdm_gui, self).close()
        if self.disable:
            return

        if n < 0:
            n = 1
        self.n += n

        delta_it = self.n - self.last_print_n  # should be n?
        if delta_it >= self.miniters:
            # We check the counter first, to reduce the overhead of time()
            cur_t = time()
            delta_t = cur_t - self.last_print_t
            if delta_t >= self.mininterval:
                elapsed = cur_t - self.start_t
                # EMA (not just overall average)
                if self.smoothing and delta_t:
                    self.avg_time = delta_t / delta_it \
                        if self.avg_time is None \
                        else self.smoothing * delta_t / delta_it + \
                        (1 - self.smoothing) * self.avg_time

                # Inline due to multiple calls
                total = self.total
                ax = self.ax

                # instantaneous rate
                y = delta_it / delta_t
                # smoothed rate
                z = self.n / elapsed
                # update line data
                self.xdata.append(self.n * 100.0 / total
                                  if total else cur_t)
                self.ydata.append(y)
                self.zdata.append(z)

                # Discard old values
                if (not total) and elapsed > 66:
                    self.xdata.popleft()
                    self.ydata.popleft()
                    self.zdata.popleft()

                ymin, ymax = ax.get_ylim()
                if y > ymax or z > ymax:
                    ymax = 1.1 * y
                    ax.set_ylim(ymin, ymax)
                    ax.figure.canvas.draw()

                if total:
                    self.line1.set_data(self.xdata, self.ydata)
                    self.line2.set_data(self.xdata, self.zdata)
                    try:
                        poly_lims = self.hspan.get_xy()
                    except AttributeError:
                        self.hspan = self.plt.axhspan(0, 0.001, xmin=0,
                                                      xmax=0, color='g')
                        poly_lims = self.hspan.get_xy()
                    poly_lims[0, 1] = ymin
                    poly_lims[1, 1] = ymax
                    poly_lims[2] = [self.n / total, ymax]
                    poly_lims[3] = [poly_lims[2, 0], ymin]
                    if len(poly_lims) > 4:
                        poly_lims[4, 1] = ymin
                    self.hspan.set_xy(poly_lims)
                else:
                    t_ago = [cur_t - i for i in self.xdata]
                    self.line1.set_data(t_ago, self.ydata)
                    self.line2.set_data(t_ago, self.zdata)

                ax.set_title(self.format_meter(
                    self.n, total, elapsed, 0,
                    self.desc, self.ascii, self.unit, self.unit_scale,
                    1 / self.avg_time if self.avg_time else None,
                    self.bar_format),
                    fontname="DejaVu Sans Mono", fontsize=11)
                self.plt.pause(1e-9)

                # If no `miniters` was specified, adjust automatically to the
                # maximum iteration rate seen so far.
                # e.g.: After running `tqdm.update(5)`, subsequent
                # calls to `tqdm.update()` will only cause an update after
                # at least 5 more iterations.
                if self.dynamic_miniters:
                    if self.maxinterval and delta_t > self.maxinterval:
                        self.miniters = self.miniters * self.maxinterval \
                            / delta_t
                    elif self.mininterval and delta_t:
                        self.miniters = self.smoothing * delta_it \
                            * self.mininterval / delta_t + \
                            (1 - self.smoothing) * self.miniters
                    else:
                        self.miniters = self.smoothing * delta_it + \
                            (1 - self.smoothing) * self.miniters

                # Store old values for next call
                self.last_print_n = self.n
                self.last_print_t = cur_t

    def close(self):
        # if not self.gui:
        #   return super(tqdm_gui, self).close()
        if self.disable:
            return

        self.disable = True

        self._instances.remove(self)

        # Restore toolbars
        self.mpl.rcParams['toolbar'] = self.toolbar
        # Return to non-interactive mode
        if not self.wasion:
            self.plt.ioff()
        if not self.leave:
            self.plt.close(self.fig)


def tgrange(*args, **kwargs):
    """
    A shortcut for tqdm_gui(xrange(*args), **kwargs).
    On Python3+ range is used instead of xrange.
    """
    return tqdm_gui(_range(*args), **kwargs)



# Copyright (c) 2017 The Regents of the University of Michigan
# All rights reserved.
# This software is licensed under the BSD 3-Clause License.
"""Workflow management based on the signac framework.

The signac-flow package provides the basic infrastructure to easily
configure and implement a workflow to operate on a signac_ data space.

.. _signac: https://glotzerlab.engin.umich.edu/signac
"""
from __future__ import absolute_import
from . import environment
from . import scheduling
from . import errors
from .project import FlowProject
from .project import JobOperation
from .project import label
from .project import classlabel
from .project import staticlabel
from .operations import cmd
from .operations import directives
from .operations import run
from .environment import get_environment
from .template import init
from .util.misc import redirect_log

__version__ = '0.6.3'

__all__ = [
    'environment',
    'scheduling',
    'errors',
    'FlowProject',
    'JobOperation',
    'label',
    'classlabel',
    'staticlabel',
    'cmd',
    'directives',
    'run',
    'init',
    'redirect_log',
    'get_environment',
    ]



# Copyright (c) 2018 The Regents of the University of Michigan
# All rights reserved.
# This software is licensed under the BSD 3-Clause License.
"""Implementation of the scheduling system for SLURM schedulers.

This module implements the Scheduler class, and the JobCluster class
for SLURM schedulers.
"""
from __future__ import print_function
import getpass
import subprocess
import tempfile
import logging
import errno

from .base import Scheduler
from .base import ClusterJob, JobStatus


logger = logging.getLogger(__name__)


def _fetch(user=None):
    "Fetch the cluster job status information from the SLURM scheduler."

    def parse_status(s):
        s = s.strip()
        if s == 'PD':
            return JobStatus.queued
        elif s == 'R':
            return JobStatus.active
        elif s in ['CG', 'CD', 'CA', 'TO']:
            return JobStatus.inactive
        elif s in ['F', 'NF']:
            return JobStatus.error
        return JobStatus.registered

    if user is None:
        user = getpass.getuser()

    cmd = ['squeue', '-u', user, '-h', "--format=%2t%100j"]
    try:
        result = subprocess.check_output(cmd).decode('utf-8')
    except subprocess.CalledProcessError as error:
        print('error', error)
        raise
    except IOError as error:
        if error.errno != errno.ENOENT:
            raise
        else:
            raise RuntimeError("Slurm not available.")
    lines = result.split('\n')
    for line in lines:
        if line:
            status = line[:2]
            name = line[2:].rstrip()
            yield ClusterJob(name, parse_status(status))


class SlurmJob(ClusterJob):
    "A Slurm is a ClusterJob managed by a SLURM scheduler."
    pass


class SlurmScheduler(Scheduler):
    """Implementation of the abstract Scheduler class for TORQUE schedulers.

    This class allows us to submit cluster jobs to a TORQUE scheduler and query
    their current status.

    :param user:
        Limit the status information to cluster jobs submitted by user.
    :type user:
        str
    """
    # The standard command used to submit jobs to the SLURM scheduler.
    submit_cmd = ['sbatch']

    def __init__(self, user=None, **kwargs):
        super(SlurmScheduler, self).__init__(**kwargs)
        self.user = user

    def jobs(self):
        "Yield cluster jobs by querying the scheduler."
        self._prevent_dos()
        for job in _fetch(user=self.user):
            yield job

    def submit(self, script, after=None, hold=False, pretend=False, flags=None, **kwargs):
        """Submit a job script for execution to the scheduler.

        :param script:
            The job script submitted for execution.
        :type script:
            str
        :param after:
            Execute the submitted script after a job with this id has completed.
        :type after:
            str
        :param pretend:
            If True, do not actually submit the script, but only simulate the submission.
            Can be used to test whether the submission would be successful.
            Please note: A successful "pretend" submission is not guaranteed to succeeed.
        :type pretend:
            bool
        :param flags:
            Additional arguments to pass through to the scheduler submission command.
        :type flags:
            list
        :returns:
            Returns True if the cluster job was successfully submitted, otherwise None.
        """
        if flags is None:
            flags = []
        elif isinstance(flags, str):
            flags = flags.split()

        submit_cmd = self.submit_cmd + flags

        if after is not None:
            submit_cmd.extend(
                ['-W', 'depend="afterany:{}"'.format(after.split('.')[0])])

        if hold:
            submit_cmd += ['--hold']

        if pretend:
            print("# Submit command: {}".format('  '.join(submit_cmd)))
            print(script)
            print()
        else:
            with tempfile.NamedTemporaryFile() as tmp_submit_script:
                tmp_submit_script.write(str(script).encode('utf-8'))
                tmp_submit_script.flush()
                subprocess.check_output(submit_cmd + [tmp_submit_script.name])
                return True

    @classmethod
    def is_present(cls):
        "Return True if it appears that a SLURM scheduler is available within the environment."
        try:
            subprocess.check_output(['sbatch', '--version'], stderr=subprocess.STDOUT)
        except (IOError, OSError):
            return False
        else:
            return True



# Copyright (c) 2018 The Regents of the University of Michigan
# All rights reserved.
# This software is licensed under the BSD 3-Clause License.
"""Provide jinja2 template environment filter functions."""


def _identical(iterable):
    """Check that all elements of an iterator are identical"""
    return len(set(iterable)) <= 1


def _format_timedelta(delta):
    "Format a time delta for interpretation by schedulers."
    if isinstance(delta, int) or isinstance(delta, float):
        import datetime
        delta = datetime.timedelta(hours=delta)
    hours, r = divmod(delta.seconds, 3600)
    minutes, seconds = divmod(r, 60)
    hours += delta.days * 24
    return "{:0>2}:{:0>2}:{:0>2}".format(hours, minutes, seconds)


def _with_np_offset(operations):
    """Add the np_offset variable to the operations' directives."""
    offset = 0
    for operation in operations:
        operation.directives.setdefault('np_offset', offset)
        offset += operation.directives['np']
    return operations



# Copyright (c) 2018 The Regents of the University of Michigan
# All rights reserved.
# This software is licensed under the BSD 3-Clause License.
"Simple progressbar formatting."
import sys


def with_progressbar(iterable, total=None, width=120, desc='',
                     percentage=True, file=None):
    if file is None:
        file = sys.stderr
    if total is None:
        total = len(iterable)
    n = max(1, total // width)

    left = desc + '|'
    right = '|{p:>4.0%}' if percentage else '|'
    w = width - len(left) - len(right)

    def _draw(p):
        f = int(p*w)
        n = w-f
        if p:
            file.write('\r' * width)
        bar = left + '#' * f + '-' * n + right
        file.write(bar.format(p=p))
        file.flush()

    try:
        for i, item in enumerate(iterable):
            if i % n == 0:
                _draw(i/total)
            yield item
        _draw(1.0)
    finally:
        file.write('\n')
        file.flush()


__all__ = ['with_progressbar']



import os
import sys
import argparse
import time
import uuid
import shutil
import threading
import json
import logging
from contextlib import contextmanager

from signac import Collection
from flow.scheduling.base import JobStatus


logger = logging.getLogger(__name__)


FMT_STATUS = {
    JobStatus.inactive: 'I',
    JobStatus.queued: 'Q',
    JobStatus.active: 'A',
}


def _get_submit_parser(parser_submit=None):
    if parser_submit is None:
        parser_submit = argparse.ArgumentParser()
    parser_submit.add_argument(
        '--job-name',
        required=True,
        help="The name of the job.")
    parser_submit.add_argument(
        '-D', '--chdir',
        help="Change to this directory prior to execution.")
    return parser_submit


def _get_args(script):
    for line in script.readlines():
        if line.startswith('#SSCHED'):
            for token in line[8:].split():
                yield token


def main_submit(args):
    # Try to parse args, should raise error if anytyhing is wrong or missing
    with open(args.filename) as script:
        _get_submit_parser().parse_args(list(_get_args(script)))
    dst = os.path.join(args.inbox, str(uuid.uuid4()) + '.sh')
    shutil.copyfile(args.filename, dst)


@contextmanager
def _lock_database(args):
    fn_lock = args.db + '.lock'
    remove_lock = True
    try:
        with open(fn_lock, mode='x'):
            with Collection.open(args.db) as db:
                yield db
    except FileExistsError:
        remove_lock = False
        raise RuntimeError(
            "Unable to lock database file '{}'. Is another process already "
            "running?".format(args.db))
    finally:
        if remove_lock:
            try:
                os.unlink(fn_lock)
            except Exception:
                pass


def _list_files_by_mtime(path):
    files = os.listdir(path)
    return sorted(files, key=lambda fn: os.path.getmtime(os.path.join(path, fn)))


def _process_inbox_loop(inbox, queue, db, stop):
    parser = _get_submit_parser()

    while True:
        logger.info("Processing inbox...")
        for fn in _list_files_by_mtime(args.inbox):
            src = os.path.join(inbox, fn)
            dst = os.path.join(queue, fn)
            os.rename(src, dst)
            with open(dst) as script:
                submit_args = parser.parse_args(list(_get_args(script)))
            doc = vars(submit_args)
            doc['script'] = dst
            doc['status'] = int(JobStatus.queued)
            doc['_queued'] = time.time()
            _id = db.insert_one(doc)
            logger.info("Queued '{}'.".format(_id))
        db.flush()
        if stop.wait(timeout=5):
            break


def _process_queue(args, db):
    from subprocess import check_call, CalledProcessError
    logger.info("Processing queue...")
    db.delete_many({'_delete_after.$lt': time.time()})

    docs = db.find({'status': int(JobStatus.queued)})
    for doc in sorted(docs, key=lambda doc: doc['_id']):
        break
    else:
        logger.info("No jobs...")
        return

    doc['status'] = int(JobStatus.active)
    db[doc['_id']] = doc
    db.flush()

    cwd = os.getcwd()
    try:
        logger.info("Executing job '{}' ({})...".format(doc['job_name'], doc['_id']))
        try:
            chdir = doc.get('chdir', '').strip('"') or os.path.expanduser('~')
            os.chdir(chdir)
        except FileNotFoundError as error:
            logger.warning(error)
        for i in range(100):
            fn_out = '{:05x}.out.{}'.format(int(doc['_id'], 16), i)
            if not os.path.exists(fn_out):
                break
        else:
            assert False    # This point should not be reached.
        for i in range(100):
            fn_err = '{:05x}.err.{}'.format(int(doc['_id'], 16), i)
            if not os.path.exists(fn_err):
                break
        else:
            assert 0

        cmd = '/bin/bash ' + doc['script']

        with open(fn_out, 'w') as outfile:
            with open(fn_err, 'w') as errfile:
                check_call(cmd, shell=True, stdout=outfile, stderr=errfile)
    except CalledProcessError as error:
        logger.warning("Error while executing job '{}'.".format(doc['_id']))
    finally:
        os.chdir(cwd)
        os.remove(doc['script'])
        doc['status'] = int(JobStatus.inactive)
        doc['_delete_after'] = time.time() + 2 * 60     # remove after 2 mins
        db[doc['_id']] = doc
        db.flush()


def main_run(args):
    print("Start scheduler...")
    os.makedirs(args.inbox, exist_ok=True)
    os.makedirs(args.queue, exist_ok=True)
    with _lock_database(args) as db:
        print("Execute this to enable environment detection for this scheduler:")
        print('export SIMPLE_SCHEDULER="{} --data={}"'.format(sys.argv[0], args.data))
        stop = threading.Event()
        process_inbox_thread = threading.Thread(
            target=_process_inbox_loop,
            kwargs=dict(inbox=args.inbox, queue=args.queue, db=db, stop=stop))
        try:
            process_inbox_thread.start()
            while True:
                _process_queue(args, db)
                time.sleep(5)
        except KeyboardInterrupt:
            stop.set()
            process_inbox_thread.join()
            print("Stopping...")


def main_status(args):
    with Collection.open(args.db, mode='r') as db:
        if args.json:
            print(json.dumps({doc['_id']: doc for doc in db}, indent=4))
        else:
            for _id in sorted(db.ids):
                doc = db[_id]
                print('{}\t{}\t{}'.format(
                    doc['_id'],
                    doc['job_name'],
                    FMT_STATUS[JobStatus(doc['status'])]))


if __name__ == '__main__':
    default_home_data = os.environ.get(
        'SIMPLE_SCHEDULER_DATA_HOME',
        os.path.expanduser('~/.local/share/simple-scheduler'))

    parser = argparse.ArgumentParser()
    parser.add_argument(
        '--data',
        default=default_home_data,
        help="Specify the path to the scheduler's data directory. "
             "Default: {}".format(default_home_data))
    parser.add_argument(
        '--max-queue-size',
        type=int,
        default=10000,
        help="Specify the maximum number of jobs that can be queued.")

    subparsers = parser.add_subparsers()

    parser_submit = subparsers.add_parser(
        'submit',
        description="Submit a job script for execution to the scheduler.")
    parser_submit.add_argument(
        'filename',
        help="The path to the script to submit for execution.")
    parser_submit.set_defaults(func=main_submit)

    parser_status = subparsers.add_parser(
        'status',
        description="Display the status of jobs submitted to the scheduler.")
    parser_status.add_argument('--json', action='store_true')
    parser_status.set_defaults(func=main_status)

    parser_run = subparsers.add_parser(
        'run',
        description="Execute jobs subitted to the schduler.")
    parser.add_argument(
        '--polling-period',
        default=10,
        help="Check the database for new entryies every given seconds.")
    parser_run.set_defaults(func=main_run)

    args = parser.parse_args()
    if not hasattr(args, 'func'):
        parser.print_usage()
        sys.exit(2)

    args.inbox = os.path.join(args.data, 'inbox')
    args.queue = os.path.join(args.data, 'queue')
    args.db = os.path.join(args.data, 'db.txt')

    logging.basicConfig(level=logging.INFO)
    args.func(args)



# Copyright (c) 2018 The Regents of the University of Michigan
# All rights reserved.
# This software is licensed under the BSD 3-Clause License.
"""This module implements the run() function, which when called equips
a regular Python module with a command line interface, which can be used
to execute functions defined within the same module, that operate on a
signac data space.
"""
from __future__ import print_function
import sys
import argparse
import logging
import inspect
from multiprocessing import Pool

from signac import get_project
from signac.common import six

from .util.tqdm import tqdm
from .util.execution import fork


logger = logging.getLogger(__name__)


def cmd(func):
    """Specifies that ``func`` returns a shell command.

    If this function is an operation function defined by :class:`~.FlowProject`, it will
    be interpreted to return a shell command, instead of executing the function itself.

    For example:

    .. code-block:: python

        @FlowProject.operation
        @flow.cmd
        def hello(job):
            return "echo {job._id}"
    """
    setattr(func, '_flow_cmd', True)
    return func


class directives(object):
    """Decorator for operation functions to provide additional execution directives.

    Directives can for example be used to provide information about required resources
    such as the number of processes required for execution of parallelized operations.
    """

    def __init__(self, **kwargs):
        self.kwargs = kwargs

    @classmethod
    def copy_from(cls, func):
        return cls(** getattr(func, '_flow_directives', dict()))

    def __call__(self, func):
        directives = getattr(func, '_flow_directives', dict())
        directives.update(self.kwargs)
        setattr(func, '_flow_directives', directives)
        return func


def _get_operations(include_private=False):
    """"Yields the name of all functions that qualify as an operation function.

    The module is inspected and all functions that have only one argument
    is yielded. Unless the 'include_private' argument is True, all private
    functions, that means the name starts with one or more '_' characters
    are ignored.
    """
    module = inspect.getmodule(inspect.currentframe().f_back.f_back)
    for name, obj in inspect.getmembers(module):
        if not include_private and name.startswith('_'):
            continue
        if inspect.isfunction(obj):
            if six.PY2:
                signature = inspect.getargspec(obj)
            else:
                signature = inspect.getfullargspec(obj)
            if len(signature.args) == 1:
                yield name


def run(parser=None):
    """Access to the "run" interface of an operations module.

    Executing this function within a module will start a command line interface,
    that can be used to execute operations defined within the same module.
    All **top-level unary functions** will be intepreted as executable operation functions.

    For example, if we have a module as such:

    .. code-block:: python

        # operations.py

        def hello(job):
            print('hello', job)

        if __name__ == '__main__':
            import flow
            flow.run()

    Then we can execute the ``hello`` operation for all jobs from the command like like this:

    .. code-block:: bash

        $ python operations.py hello

    .. note::

        You can control the degree of parallelization with the ``--np`` argument.


    For more information, see:

    .. code-block:: bash

        $ python operations.py --help
    """
    if parser is None:
        parser = argparse.ArgumentParser()

    parser.add_argument(
        'operation',
        type=str,
        choices=list(_get_operations()),
        help="The operation to execute.")
    parser.add_argument(
        'jobid',
        type=str,
        nargs='*',
        help="The job ids, as registered in the signac project. "
             "Omit to default to all statepoints.")
    parser.add_argument(
        '--np',
        type=int,
        default=1,
        help="Specify the number of cores to parallelize to (default=1) or 0 "
             "to parallelize on as many cores as there are available.")
    parser.add_argument(
        '-t', '--timeout',
        type=int,
        help="A timeout in seconds after which the parallel execution "
             "of operations is canceled.")
    parser.add_argument(
        '--progress',
        action='store_true',
        help="Display a progress bar during execution.")
    args = parser.parse_args()

    project = get_project()

    def _open_job_by_id(_id):
        try:
            return project.open_job(id=_id)
        except KeyError:
            msg = "Did not find job corresponding to id '{}'.".format(_id)
            raise KeyError(msg)
        except LookupError:
            raise LookupError("Multiple matches for id '{}'.".format(_id))

    if len(args.jobid):
        try:
            jobs = [_open_job_by_id(jid) for jid in args.jobid]
        except (KeyError, LookupError) as e:
            print(e, file=sys.stderr)
            sys.exit(1)
    else:
        jobs = project

    module = inspect.getmodule(inspect.currentframe().f_back)
    try:
        operation_func = getattr(module, args.operation)
    except AttributeError:
        raise KeyError("Unknown operation '{}'.".format(args.operation))

    if getattr(operation_func, '_flow_cmd', False):

        def operation(job):
            cmd = operation_func(job).format(job=job)
            fork(cmd=cmd, timeout=args.timeout)
    else:
        operation = operation_func

    # Serial execution
    if args.np == 1 or len(jobs) < 2:
        if args.timeout is not None:
            logger.warning("A timeout has no effect in serial execution!")
        for job in tqdm(jobs) if args.progress else jobs:
            operation(job)

    # Parallel execution
    elif six.PY2:
        # Due to Python 2.7 issue #8296 (http://bugs.python.org/issue8296) we
        # always need to provide a timeout to avoid issues with "hanging"
        # processing pools.
        timeout = sys.maxint if args.timeout is None else args.timeout
        pool = Pool(args.np)
        result = pool.imap_unordered(operation, jobs)
        for _ in tqdm(jobs) if args.progress else jobs:
            result.next(timeout)
    else:
        with Pool(args.np) as pool:
            result = pool.imap_unordered(operation, jobs)
            for _ in tqdm(jobs) if args.progress else jobs:
                result.next(args.timeout)


__all__ = ['cmd', 'directives', 'run']



# Copyright (c) 2018 The Regents of the University of Michigan
# All rights reserved.
# This software is licensed under the BSD 3-Clause License.
"Helper functions to keep track of abbreviations meant for output on screen."


def abbreviate(x, a):
    "Abbreviate x with a and add to the abbreviation table."
    if x == a:
        return x
    else:
        abbreviate.table[a] = x
        return a

abbreviate.table = dict()  # noqa


def shorten(x, max_length=None):
    "Shorten x to max_length and add to abbreviation table."
    if max_length is None:
        return x
    else:
        return abbreviate(x, x[:max_length])



# Copyright (c) 2018 The Regents of the University of Michigan
# All rights reserved.
# This software is licensed under the BSD 3-Clause License.
"""Implementation of the scheduling system for a fake scheduler.

The FakeScheduler class can be used in place of a real scheduler to test
the cluster job submission workflow.
"""
from __future__ import print_function
import logging

from .base import Scheduler

logger = logging.getLogger(__name__)


class FakeScheduler(Scheduler):
    """Implementation of the abstract Scheduler class for a fake scheduler.

    This scheduler does not actually schedule (or execute) any jobs, but it can be used
    to test the submission workflow.
    """

    def jobs(self):
        "Yields nothing, since the FakeScheduler does not actually schedule any jobs."
        return
        yield

    def submit(self, script, **kwargs):
        "Just print the script to screen."
        print(script)

    @classmethod
    def is_present(cls):
        return False



# Copyright (c) 2018 The Regents of the University of Michigan
# All rights reserved.
# This software is licensed under the BSD 3-Clause License.
"""This module contains the FlowProject module templates.

These templates can be initialized via the init() function defined
in this module and the main 'flow' command line interface.
"""
from __future__ import print_function
import os
import sys
import errno
import logging

from signac.common import six
try:
    import jinja2
    from jinja2 import TemplateNotFound as Jinja2TemplateNotFound
except ImportError:
    # Mock exception, which will never be raised.
    class Jinja2TemplateNotFound(Exception):
        pass

    JINJA2 = False
else:
    JINJA2 = True

from .util.misc import _is_identifier


logger = logging.getLogger(__name__)
if six.PY2:
    logger.addHandler(logging.NullHandler())


TEMPLATES = {
    'minimal': [('{alias}.py', 'project_minimal.pyt'), ],
    'example': [('{alias}.py', 'project_example.pyt'), ],
}


def init(alias=None, template=None, root=None, out=None):
    "Initialize a templated FlowProject module."
    if not JINJA2:
        raise ValueError("The init() function requires the 'jinja2' package.")

    if alias is None:
        alias = 'project'
    elif not _is_identifier(alias):
        raise ValueError(
            "The alias '{}' is not a valid Python identifier and therefore "
            "not be used as a FlowProject alias.".format(alias))
    if template is None:
        template = 'minimal'
    if out is None:
        out = sys.stderr

    if os.path.splitext(alias)[1]:
        raise RuntimeError("Please provide a name without suffix!")

    project_class_name = alias.capitalize()
    if not project_class_name.endswith('Project'):
        project_class_name += 'Project'

    template_environment = jinja2.Environment(
        loader=jinja2.ChoiceLoader([
            jinja2.FileSystemLoader('templates'),
            jinja2.PackageLoader('flow', 'templates')]),
        trim_blocks=True)

    context = dict()
    context['alias'] = alias
    context['project_class_name'] = project_class_name

    # render all templates
    codes = dict()

    for fn, fn_template in TEMPLATES[template]:
        fn_ = fn.format(alias=alias)   # some of the filenames may depend on the alias
        template = template_environment.get_template(fn_template)
        codes[fn_] = template.render(** context)

    # create files
    files_created = []
    for fn, code in codes.items():
        try:
            if root is not None:
                fn = os.path.join(root, fn)
            if six.PY2:
                # Adapted from: http://stackoverflow.com/questions/10978869/
                flags = os.O_CREAT | os.O_WRONLY | os.O_EXCL
                fd = os.open(fn, flags)
                with os.fdopen(fd, 'w') as file:
                    file.write(code + '\n')
            else:
                with open(fn, 'x') as fw:
                    fw.write(code + '\n')
        except OSError as e:
            if e.errno == errno.EEXIST:
                logger.error(
                    "Error while trying to initialize flow project with alias '{alias}', "
                    "a file named '{fn}' already exists!".format(alias=alias, fn=fn))
            else:
                logger.error(
                    "Error while trying to initialize flow project with alias '{alias}': "
                    "'{error}'.".format(alias=alias, error=e))
        else:
            files_created.append(fn)
            print("Created file '{}'.".format(fn), file=out)
    return files_created



# Copyright (c) 2018 The Regents of the University of Michigan
# All rights reserved.
# This software is licensed under the BSD 3-Clause License.
"""Detection of compute environments.

This module provides the ComputeEnvironment class, which can be
subclassed to automatically detect specific computational environments.

This enables the user to adjust their workflow based on the present
environment, e.g. for the adjustemt of scheduler submission scripts.
"""
from __future__ import print_function
from __future__ import division
import os
import sys
import re
import socket
import logging
import warnings
import io
import importlib
from math import ceil
from collections import OrderedDict

from signac.common import config
from signac.common import six
from signac.common.six import with_metaclass

from .scheduling.base import JobStatus
from .scheduling.slurm import SlurmScheduler
from .scheduling.torque import TorqueScheduler
from .scheduling.simple_scheduler import SimpleScheduler
from .scheduling.fakescheduler import FakeScheduler
from .util import config as flow_config
from .errors import SubmitError
from .errors import NoSchedulerError

if six.PY2:
    import imp
else:
    import importlib.machinery

logger = logging.getLogger(__name__)
if six.PY2:
    logger.addHandler(logging.NullHandler())


# Global variable can be used to override detected environment
ENVIRONMENT = None


NUM_NODES_WARNING = """Unable to determine the reqired number of nodes (nn) for this submission.
Either provide this value directly with '--nn' or provide the number of processors
per node: '--ppn'.

Please note, you can ignore this message by specifying extra submission options
with '--' or by using the '--force' option."""

UTILIZATION_WARNING = """You either specified or the environment is configured to use {ppn}
processors per node (ppn), however you only use {usage:0.2%} of each node.
Consider to increase the number of processors per operation (--np)
or adjust the processors per node (--ppn).

Alternatively, you can also use --force to ignore this warning.
"""


def setup(py_modules, **attrs):
    """Setup function for environment modules.

    Use this function in place of setuptools.setup to not only install
    a environments module, but also register it with the global signac
    configuration. Once registered, is automatically imported when the
    get_environment() function is called.
    """
    import setuptools
    from setuptools.command.install import install

    class InstallAndConfig(install):

        def run(self):
            super(InstallAndConfig, self).run()
            cfg = config.read_config_file(config.FN_CONFIG)
            try:
                envs = cfg['flow'].as_list('environment_modules')
            except KeyError:
                envs = []
            new = set(py_modules).difference(envs)
            if new:
                for name in new:
                    self.announce(
                        msg="registering module '{}' in global signac configuration".format(name),
                        level=2)
                cfg.setdefault('flow', dict())
                cfg['flow']['environment_modules'] = envs + list(new)
                cfg.write()

    return setuptools.setup(
        py_modules=py_modules,
        cmdclass={'install': InstallAndConfig},
        **attrs)


def format_timedelta(delta):
    "Format a time delta for interpretation by schedulers."
    hours, r = divmod(delta.seconds, 3600)
    minutes, seconds = divmod(r, 60)
    hours += delta.days * 24
    return "{:0>2}:{:0>2}:{:0>2}".format(hours, minutes, seconds)


class ComputeEnvironmentType(type):
    """Meta class for the definition of ComputeEnvironments.

    This meta class automatically registers ComputeEnvironment definitions,
    which enables the automatic determination of the present environment.
    """

    def __init__(cls, name, bases, dct):
        if not hasattr(cls, 'registry'):
            cls.registry = OrderedDict()
        else:
            cls.registry[name] = cls
        return super(ComputeEnvironmentType, cls).__init__(name, bases, dct)


class JobScript(io.StringIO):
    """"Simple StringIO wrapper for the creation of job submission scripts.

    Using this class to write a job submission script allows us to use
    environment specific expressions, for example for MPI commands.
    """
    eol = '\n'

    def __init__(self, env):
        self._env = env
        super(JobScript, self).__init__()

    def __str__(self):
        self.seek(0)
        return self.read()

    def write(self, s):
        if six.PY2:
            super(JobScript, self).write(unicode(s))  # noqa
        else:
            super(JobScript, self).write(s)

    def writeline(self, line=''):
        "Write one line to the job script."
        self.write(line + self.eol)

    def write_cmd(self, cmd, bg=False, np=None):
        """Write a command to the jobscript.

        This command wrapper function is a convenience function, which
        adds mpi and other directives whenever necessary.

        :param cmd: The command to write to the jobscript.
        :type cmd: str
        :param np: The number of processors required for execution.
        :type np: int
        """
        if np is not None:
            warnings.warn(DeprecationWarning("Do not provide np with write_cmd()!"))
            if np > 1:
                cmd = self._env.mpi_cmd(cmd, np=np)
        if bg:
            cmd = self._env.bg(cmd)
        self.writeline(cmd)


class ComputeEnvironment(with_metaclass(ComputeEnvironmentType)):
    """Define computational environments.

    The ComputeEnvironment class allows us to automatically determine
    specific environments in order to programatically adjust workflows
    in different environments.

    The default method for the detection of a specific environemnt is to
    provide a regular expression matching the environment's hostname.
    For example, if the hostname is my_server.com, one could identify the
    environment by setting the hostname_pattern to 'my_server'.
    """
    scheduler_type = None
    hostname_pattern = None
    submit_flags = None
    template = 'base_script.sh'

    @classmethod
    def script(cls, **kwargs):
        """Return a JobScript instance.

        Derived ComputeEnvironment classes may require additional
        arguments for the creation of a job submission script.
        """
        return JobScript(cls)

    @classmethod
    def is_present(cls):
        """Determine whether this specific compute environment is present.

        The default method for environment detection is trying to match a
        hostname pattern or delegate the environment check to the associated
        scheduler type.
        """
        if cls.hostname_pattern is None:
            if cls.scheduler_type is None:
                return False
            else:
                return cls.scheduler_type.is_present()
        else:
            return re.match(
                cls.hostname_pattern, socket.gethostname()) is not None

    @classmethod
    def get_scheduler(cls):
        """Return a environment specific scheduler driver.

        The returned scheduler class provides a standardized interface to
        different scheduler implementations.
        """
        try:
            return getattr(cls, 'scheduler_type')()
        except (AttributeError, TypeError):
            raise NoSchedulerError(
                "No scheduler defined for environment '{}'.".format(cls.__name__))

    @classmethod
    def submit(cls, script, flags=None, *args, **kwargs):
        """Submit a job submission script to the environment's scheduler.

        Scripts should be submitted to the environment, instead of directly
        to the scheduler to allow for environment specific post-processing.
        """
        if flags is None:
            flags = []
        env_flags = getattr(cls, 'submit_flags', [])
        if env_flags:
            flags.extend(env_flags)

        # Hand off the actual submission to the scheduler
        if isinstance(script, JobScript):  # api version < 6
            script = str(script)

        if cls.get_scheduler().submit(script, flags=flags, *args, **kwargs):
            return JobStatus.submitted

    @staticmethod
    def bg(cmd):
        "Wrap a command (cmd) to be executed in the background."
        return cmd + ' &'

    @classmethod
    def add_args(cls, parser):
        """Add arguments related to this compute environment to an argument parser.

        :param parser:
            The argument parser to add arguments to.
        :type parser:
            :class:`argparse.ArgumentParser`
        """
        return

    @classmethod
    def get_config_value(cls, key, default=flow_config._GET_CONFIG_VALUE_NONE):
        """Request a value from the user's configuration.

        This method should be used whenever values need to be provided
        that are specific to a users's environment. A good example are
        account names.

        When a key is not configured and no default value is provided,
        a :py:class:`~.errors.SubmitError` will be raised and the user will
        be prompted to add the missing key to their configuration.

        Please note, that the key will be automatically expanded to
        be specific to this environment definition. For example, a
        key should be 'account', not 'MyEnvironment.account`.

        :param key: The environment specific configuration key.
        :type key: str
        :param default: A default value in case the key cannot be found
            within the user's configuration.
        :type key: str
        :return: The value or default value.
        :raises SubmitError: If the key is not in the user's configuration
            and no default value is provided.
        """
        return flow_config.require_config_value(key, ns=cls.__name__, default=default)


class StandardEnvironment(ComputeEnvironment):
    "This is a default environment, which is always present."

    @classmethod
    def is_present(cls):
        return True

    @classmethod
    def mpi_cmd(cls, cmd, np):
        return 'mpirun -np {np} {cmd}'.format(np=np, cmd=cmd)


class UnknownEnvironment(StandardEnvironment):
    "Deprecated 'standard' environment, replaced by 'StandardEnvironment.'"

    def __init__(self, *args, **kwargs):
        super(
            "The 'UnknownEnvironment' class has been replaced by the "
            "'StandardEnvironment' class.", DeprecationWarning)
        super(UnknownEnvironment, self).__init__(*args, **kwargs)


class TestEnvironment(ComputeEnvironment):
    """This is a test environment.

    The test environment will print a mocked submission script
    and submission commands to screen. This enables testing of
    the job submission script generation in environments without
    a real scheduler.
    """
    scheduler_type = FakeScheduler

    @classmethod
    def mpi_cmd(cls, cmd, np):
        return 'mpirun -np {np} {cmd}'.format(np=np, cmd=cmd)

    @classmethod
    def script(cls, **kwargs):
        js = super(TestEnvironment, cls).script(**kwargs)
        for key in sorted(kwargs):
            js.writeline('#TEST {}={}'.format(key, kwargs[key]))
        return js


class SimpleSchedulerEnvironment(ComputeEnvironment):
    "An environment for the simple-scheduler scheduler."
    scheduler_type = SimpleScheduler
    template = 'simple_scheduler.sh'


class TorqueEnvironment(ComputeEnvironment):
    "An environment with TORQUE scheduler."
    scheduler_type = TorqueScheduler
    template = 'torque.sh'


class MoabEnvironment(ComputeEnvironment):
    """"An environment with TORQUE scheduler.

    This class is deprecated and only kept for backwards
    compatibility.
    """
    scheduler_type = TorqueScheduler

    def __init__(self, *args, **kwargs):
        raise RuntimeError("The MoabEnvironment has been replaced by the TorqueEnvironment.")


class SlurmEnvironment(ComputeEnvironment):
    "An environment with SLURM scheduler."
    scheduler_type = SlurmScheduler
    template = 'slurm.sh'


class NodesEnvironment(ComputeEnvironment):
    """A compute environment consisting of multiple compute nodes.

    Each compute node is assumed to have a specific number of compute units, e.g., CPUs.
    """

    @classmethod
    def calc_num_nodes(cls, np_total, ppn, force=False, **kwargs):
        if ppn is None:
            try:
                ppn = getattr(cls, 'cores_per_node')
            except AttributeError:
                raise SubmitError(NUM_NODES_WARNING)

        # Calculate the total number of required nodes
        nn = int(ceil(np_total / ppn))

        if not force:  # Perform basic check concerning the node utilization.
            usage = np_total / nn / ppn
            if usage < 0.9:
                print(UTILIZATION_WARNING.format(ppn=ppn, usage=usage), file=sys.stderr)
                raise SubmitError("Bad node utilization!")
        return nn

    @classmethod
    def add_args(cls, parser):
        super(NodesEnvironment, cls).add_args(parser)
        parser.add_argument(
            '--nn',
            type=int,
            help="Specify the total number of nodes. This value is computed automatically "
                 "from the 'np' directive otherwise.")


class DefaultTorqueEnvironment(NodesEnvironment, TorqueEnvironment):
    "A default environment for environments with TORQUE scheduler."

    @classmethod
    def mpi_cmd(cls, cmd, np):
        return 'mpirun -np {np} {cmd}'.format(np=np, cmd=cmd)

    @classmethod
    def add_args(cls, parser):
        super(DefaultTorqueEnvironment, cls).add_args(parser)
        parser.add_argument(
            '-w', '--walltime',
            type=float,
            help="The wallclock time in hours.")
        parser.add_argument(
            '--hold',
            action='store_true',
            help="Submit jobs, but put them on hold.")
        parser.add_argument(
            '--after',
            type=str,
            help="Schedule this job to be executed after "
                 "completion of a cluster job with this id.")
        parser.add_argument(
            '--no-copy-env',
            action='store_true',
            help="Do not copy current environment variables into compute node environment.")

    @classmethod
    def gen_tasks(cls, js, np_total):
        """Helper function to generate the number of tasks (for overriding)"""
        js.writeline('#PBS -l nodes={}'.format(np_total))
        return js

    @classmethod
    def script(cls, _id, np_total, walltime=None, no_copy_env=False, **kwargs):
        js = super(DefaultTorqueEnvironment, cls).script()
        js.writeline('#PBS -N {}'.format(_id))
        js = cls.gen_tasks(js)
        if walltime is not None:
            js.writeline('#PBS -l walltime={}'.format(format_timedelta(walltime)))
        if not no_copy_env:
            js.writeline('#PBS -V')
        return js


class DefaultSlurmEnvironment(NodesEnvironment, SlurmEnvironment):
    "A default environment for environments with SLURM scheduler."

    @classmethod
    def mpi_cmd(cls, cmd, np):
        return 'mpirun -np {np} {cmd}'.format(np=np, cmd=cmd)

    @classmethod
    def gen_tasks(cls, js, np_total):
        """Helper function to generate the number of tasks (for overriding)"""
        js.writeline('#SBATCH --ntasks={}'.format(np_total))
        return js

    @classmethod
    def script(cls, _id, np_total, walltime=None, **kwargs):
        js = super(DefaultSlurmEnvironment, cls).script()
        js.writeline('#!/bin/bash')
        js.writeline('#SBATCH --job-name="{}"'.format(_id))
        js = cls.gen_tasks(js, np_total)
        if walltime is not None:
            js.writeline('#SBATCH -t {}'.format(format_timedelta(walltime)))
        return js

    @classmethod
    def add_args(cls, parser):
        super(DefaultSlurmEnvironment, cls).add_args(parser)
        parser.add_argument(
            '-w', '--walltime',
            type=float,
            default=12,
            help="The wallclock time in hours.")
        parser.add_argument(
            '--hold',
            action='store_true',
            help="Submit jobs, but put them on hold.")
        parser.add_argument(
            '--after',
            type=str,
            help="Schedule this job to be executed after "
                 "completion of a cluster job with this id.")


class CPUEnvironment(ComputeEnvironment):
    "An environment with CPUs."
    pass


class GPUEnvironment(ComputeEnvironment):
    "An environment with GPUs."
    pass


def _import_module(fn):
    if six.PY2:
        return imp.load_source(os.path.splitext(fn)[0], fn)
    else:
        return importlib.machinery.SourceFileLoader(fn, fn).load_module()


def _import_modules(prefix):
    for fn in os.path.listdir(prefix):
        if os.path.isfile(fn) and os.path.splitext(fn)[1] == '.py':
            _import_module(os.path.join(prefix, fn))


def _import_configured_environments():
    cfg = config.load_config(config.FN_CONFIG)
    try:
        for name in cfg['flow'].as_list('environment_modules'):
            try:
                importlib.import_module(name)
            except ImportError as e:
                logger.warning(e)
    except KeyError:
        pass


def registered_environments(import_configured=True):
    if import_configured:
        _import_configured_environments()
    return list(ComputeEnvironment.registry.values())


def get_environment(test=False, import_configured=True):
    """Attempt to detect the present environment.

    This function iterates through all defined ComputeEnvironment
    classes in reversed order of definition and and returns the
    first EnvironmentClass where the is_present() method returns
    True.

    :param test:
        Return the TestEnvironment
    :type test:
        bool
    :returns:
        The detected environment class.
    """
    if test:
        return TestEnvironment
    else:
        # Return a globally specified environment
        if ENVIRONMENT is not None:
            return ENVIRONMENT

        # Obtain a list of all registered environments
        env_types = registered_environments(import_configured=import_configured)
        logger.debug(
            "List of registered environments:\n\t{}".format(
                '\n\t'.join((str(env.__name__) for env in env_types))))

        # Select environment based on environment variable if set.
        env_from_env_var = os.environ.get('SIGNAC_FLOW_ENVIRONMENT')
        if env_from_env_var:
            for env_type in env_types:
                if env_type.__name__ == env_from_env_var:
                    return env_type
            else:
                raise ValueError("Unknown environment '{}'.".format(env_from_env_var))

        # Select based on DEBUG flag:
        for env_type in env_types:
            if getattr(env_type, 'DEBUG', False):
                logger.debug("Select environment '{}'; DEBUG=True.".format(env_type.__name__))
                return env_type

        # Default selection:
        for env_type in reversed(env_types):
            if env_type.is_present():
                logger.debug("Select environment '{}'; is present.".format(env_type.__name__))
                return env_type

        # Otherwise, just return a standard environment
        return StandardEnvironment
